from io import BytesIO

import pytest

from phylodata.errors import ValidationError
from phylodata.parsers.parse_files import parse_beast2_config


def to_bytes_io(text: str):
    bytesio = BytesIO(bytes(text, "ascii"))
    bytesio.name = "Test"
    return bytesio


def test_simple_beast_config_is_ok():
    file = to_bytes_io("""<?xml version="1.0" encoding="UTF-8" standalone="no"?>
    <beast namespace="beast.core" required="BEAST.base v2.7.7" version="2.7">
        <data></data>
        <run></run>
    </beast>""")
    parse_beast2_config(file)


def test_missing_data_tag_fails():
    with pytest.raises(ValidationError):
        file = to_bytes_io("""<?xml version="1.0" encoding="UTF-8" standalone="no"?>
        <beast namespace="beast.core" required="BEAST.base v2.7.7" version="2.7">
            <run></run>
        </beast>""")
        parse_beast2_config(file)


def test_missing_run_tag_fails():
    with pytest.raises(ValidationError):
        file = to_bytes_io("""<?xml version="1.0" encoding="UTF-8" standalone="no"?>
        <beast namespace="beast.core" required="BEAST.base v2.7.7" version="2.7">
            <data></data>
        </beast>""")
        parse_beast2_config(file)


def test_empty_file_fails():
    with pytest.raises(ValidationError):
        file = to_bytes_io("")
        parse_beast2_config(file)


def test_non_xml_file_fails():
    with pytest.raises(ValidationError):
        file = to_bytes_io("""
        beast2_config:
            - "this is a YAML and no XML file"
        """)
        parse_beast2_config(file)


def test_missing_beast_tag_fails():
    with pytest.raises(ValidationError):
        file = to_bytes_io("""<?xml version="1.0" encoding="UTF-8" standalone="no"?>
        <some-other-tag></some-other-tag>""")
        parse_beast2_config(file)



"""Generate JSON schema for PaperWithExperiment data type."""

import msgspec

from phylodata.data_types import PaperWithExperiment

schema = msgspec.json.schema(PaperWithExperiment)

with open("../website/src/lib/schema.json", "wb") as f:
    f.write(msgspec.json.format(msgspec.json.encode(schema)))



import streamlit as st

from phylodata.data_types import EvolutionaryModel
from phylodata.modules.module import Module
from phylodata.parsers.parse_evolutionary_model import parse_evolutionary_model


class EvolutionaryModelModule(Module[EvolutionaryModel]):
    def ui(self): ...

    def validate(self): ...

    def set_dependencies(self, beast2_configuration):
        self.beast2_configuration = beast2_configuration

    def parse(self) -> EvolutionaryModel:
        st.text("Detecting the evolutionary model...")
        return parse_evolutionary_model(self.beast2_configuration)



from typing import TypeVar
from xml.etree.ElementTree import Element

T = TypeVar("T")


def get_attribute(
    xml_element: Element, attribute_name: str, default: T = None
) -> str | T:
    """Returns the value of the attribute with the given name from the XML element,
    or the default value if not found.

    This respects the BEAST 2 XML format, where an attribute can also be specified
    using a child element with the same name as the attribute or an input element with
    the name attribute set to the attribute name."""

    if attribute := xml_element.attrib.get(attribute_name):
        return attribute

    for child in xml_element:
        if child.tag == attribute_name and child.text:
            return child.text

        if (
            child.tag == "input"
            and child.attrib.get("name") == attribute_name
            and child.text
        ):
            return child.text

    return default



from io import BytesIO

from phylodata.data_types import FileType
from phylodata.parsers.parse_files import parse_other_file


def to_bytes_io(text: str):
    bytesio = BytesIO(bytes(text, "ascii"))
    bytesio.name = "Test"
    return bytesio


def test_summary_tree_is_ok():
    raw_nexus = """
    #NEXUS
    Begin taxa;
       	Dimensions ntax=3;
  		Taxlabels
 			0
 			1
            2;
    End;
    Begin trees;
        tree STATE_1 = ((0, 1), 2);
    End;
    """

    file = to_bytes_io(raw_nexus)
    parsed = parse_other_file(file)

    assert parsed.type == FileType.SUMMARY_TREE



from io import BytesIO

from phylodata.data_types import Trees
from phylodata.errors import ValidationError
from phylodata.utils.bytesio_utils import get_nexus_from_bytesio

ULTRAMETRIC_REL_THRESHOLD = 1e-6
"""The absolute relative differences of the tip ages for a tree to be ultrametric. We
use an approximate check because of numerical precision issues."""


def parse_trees(beast2_trees_file: BytesIO) -> Trees:
    """Parses a BEAST 2 trees file in the NEXUS format."""
    nexus = get_nexus_from_bytesio(beast2_trees_file)

    if not nexus.TREES or not nexus.TREES.trees:
        raise ValidationError("Invalid BEAST 2 trees file.")

    # we use the last tree as a representative
    # (the first one could be biased due to initialization.
    # optimally, we would check all trees though.)
    example_tree = nexus.TREES.trees[-1]

    return Trees(
        number_of_trees=len(nexus.TREES.trees),
        number_of_tips=get_number_of_tips(example_tree.newick),
        ultrametric=is_ultrametric(example_tree.newick),
        time_tree=True,
        rooted=True,  # BEAST 2 always produces rooted trees
        ccd1_entropy=0,  # TODO
        tree_ess=0,  # TODO
        ccd0_map_tree="",  # TODO
        hipstr_tree="",  # TODO
        leaf_to_sample_map={},  # TODO
        average_root_age=get_average_root_age(nexus.TREES.trees),
    )


def get_number_of_tips(newick_node):
    """Returns the number of tips for the given commonnexus newick tree."""
    if not newick_node._descendants:
        return 1
    else:
        return sum(map(get_number_of_tips, newick_node._descendants))


def is_ultrametric(newick_node):
    """Returns if the given commonnexus newick tree is ultrametric
    up to ULTRAMETRIC_REL_THRESHOLD."""
    tip_dates = _get_tip_times_since_origin(newick_node, 0.0)
    min_date = min(tip_dates)
    max_date = max(tip_dates)
    return (max_date - min_date) <= ULTRAMETRIC_REL_THRESHOLD * max_date


def get_average_root_age(trees):
    """Returns the average age between the root and the most recent tip for
    the given commonnexus trees."""
    root_ages = []

    for tree in trees:
        tip_dates = _get_tip_times_since_origin(tree.newick, 0.0)
        root_age = max(tip_dates) - tree.newick.length
        root_ages.append(root_age)

    return sum(root_ages) / len(root_ages)


def _get_tip_times_since_origin(newick_node, accumulated_height):
    """A helper function returning the time between the origin and the given
    commonnexus newick node."""
    if not newick_node._descendants:
        return [accumulated_height + newick_node.length]
    else:
        return [
            date
            for child in newick_node._descendants
            for date in _get_tip_times_since_origin(
                child, accumulated_height + newick_node.length
            )
        ]



import subprocess
from pathlib import Path

import click

import phylodata
from phylodata.data_types import get_schema, validate_json


@click.group()
def cli(): ...


@cli.command(help="Process an experiment before uploading it to PhyloData.")
def process():
    subprocess.run(["streamlit", "run", str(Path(phylodata.__path__[0]) / "ui.py")])  # type: ignore


@cli.command(help="Validate if a given JSON file contains valid PhyloData metadata.")
@click.argument("file_path", type=click.Path(exists=True))
def validate(file_path: str):
    validate_json(file_path)
    print("File is valid!")


@cli.command(help="Prints the JSON schema for valid PhyloData metadata files.")
def schema():
    print(get_schema())



import re
from time import sleep

import requests
from loguru import logger

from phylodata.errors import BlastError

BLAST_URL = "https://blast.ncbi.nlm.nih.gov/blast/Blast.cgi"
WAIT_TIME_S = 5
MAX_BATCH_SIZE = 32


def run_blast(
    sequences: list[str], parameters: dict, max_length_considered: int
) -> list:
    results = []

    for i in range(0, len(sequences), MAX_BATCH_SIZE):
        batch = sequences[i : i + MAX_BATCH_SIZE]

        fasta_data = build_fasta_data(batch, max_length_considered)
        parameters["QUERY"] = fasta_data

        request_id, _ = initiate_blast_request(fasta_data, parameters)
        wait_until_ready(request_id)

        results += fetch_results(request_id)["BlastOutput2"]

    return results


def initiate_blast_request(fasta_data: str, parameters: dict) -> tuple[str, int]:
    response = requests.post(BLAST_URL, data=parameters)
    if response.status_code != 200:
        logger.error("BLAST submission failed: " + response.text)
        raise BlastError("BLAST submission failed.")

    request_id = re.search(r"RID = (\S+)", response.text)
    wait_time_s = re.search(r"RTOE = (\d+)", response.text)

    if not request_id or not wait_time_s:
        logger.error("Failed to parse BLAST response: " + response.text)
        raise BlastError("Failed to parse BLAST response.")

    request_id = request_id.group(1)
    wait_time_s = int(wait_time_s.group(1))

    return request_id, wait_time_s


def build_fasta_data(sequences: list[str], max_length_considered: int) -> str:
    fasta_data = ""

    for i, seq in enumerate(sequences):
        fasta_data += f">{i}\n{seq[:max_length_considered]}\n"

    return fasta_data


def wait_until_ready(request_id: str):
    result_params = {
        "CMD": "Get",
        "RID": request_id,
    }

    response = requests.get(BLAST_URL, params=result_params)
    status = re.findall("Status=(.*)", response.text)[0]

    match status:
        case "READY":
            return
        case "WAITING":
            logger.info("Keep waiting for BLAST results...")
            sleep(WAIT_TIME_S)
            wait_until_ready(request_id)
        case _:
            logger.error(
                f"BLAST submission failed with status {status}: {response.text}"
            )
            raise Exception(f"BLAST submission failed: {status}")


def fetch_results(request_id: str) -> dict:
    result_params = {
        "CMD": "Get",
        "RID": request_id,
        "FORMAT_TYPE": "JSON2_S",
        "ALIGNMENTS": 1,
        "DESCRIPTIONS": 1,
    }

    response = requests.get(BLAST_URL, params=result_params)

    try:
        result = response.json()
    except Exception:
        logger.error(f"Failed to parse BLAST results: {response.text}")
        raise BlastError("Failed to parse BLAST results.")

    return result


def extract_taxon_ids(blast_json: list) -> list[int | None]:
    taxon_ids = []

    for result in blast_json:
        try:
            taxon_id = result["report"]["results"]["search"]["hits"][0]["description"][
                0
            ]["taxid"]
            taxon_id = int(taxon_id)
        except (KeyError, IndexError):
            taxon_id = None

        taxon_ids.append(taxon_id)

    return taxon_ids



from io import BytesIO

from phylodata.data_types import EvolutionaryModel, EvolutionaryModelComponent
from phylodata.parsers.parse_beast2_packages import BEAST2_PACKAGE_PARSERS
from phylodata.utils.bytesio_utils import get_xml_from_bytesio


def parse_evolutionary_model(beast2_config: BytesIO) -> EvolutionaryModel:
    xml = get_xml_from_bytesio(beast2_config)

    models = []

    for parser in BEAST2_PACKAGE_PARSERS:
        if parser.is_used(xml):
            model = EvolutionaryModelComponent(
                name=parser.get_name(),
                description=parser.get_description(),
                parameters=parser.get_parameters(xml),
                documentation_url=parser.get_documentation_url(),
                type=parser.get_type(),
            )
            models.append(model)

    return EvolutionaryModel(models=models)



import csv
from hashlib import md5
from io import BytesIO, TextIOWrapper
from typing import Optional

from phylodata.data_types import File, FileType
from phylodata.errors import ValidationError
from phylodata.utils.bytesio_utils import get_nexus_from_bytesio, get_xml_from_bytesio

MIN_NUM_SNAPSHOTS = 50
"""The minimum number of snapshots required for a valid BEAST 2 posterior log file.
When the number of snapshots is less than this value, the log file is considered invalid."""


def parse_file(file: BytesIO, file_type: Optional[FileType] = FileType.UNKNOWN) -> File:
    match file_type:
        case FileType.BEAST2_CONFIGURATION:
            return parse_beast2_config(file)
        case FileType.BEAST2_POSTERIOR_LOGS:
            return parse_beast2_logs(file)
        case FileType.BEAST2_POSTERIOR_TREES:
            return parse_beast2_trees(file)
        case _:
            return parse_other_file(file)


def parse_beast2_config(file: BytesIO) -> File:
    """Parses a BEAST 2 config file. If the file does not seem to be a valid
    BEAST config, a ValidationError is raised."""
    xml = get_xml_from_bytesio(file)
    root = xml.getroot()

    if root.tag.lower() != "beast":
        raise ValidationError("BEAST 2 configuration has no root BEAST tag.")

    if not any(child.tag.lower() == "data" for child in root):
        raise ValidationError("BEAST 2 has no <data> tag.")

    if not any(child.tag.lower() == "run" for child in root):
        raise ValidationError("BEAST 2 has no <run> tag.")

    buffer = file.getbuffer()
    return File(
        name=file.name,
        type=FileType.BEAST2_CONFIGURATION,
        version=1,
        size_bytes=buffer.nbytes,
        md5=md5(buffer).hexdigest(),
    )


def parse_beast2_logs(file: BytesIO) -> File:
    """Parses a BEAST 2 log file. If the file does not seem to be a valid
    BEAST log, a ValidationError is raised."""
    try:
        wrapper = TextIOWrapper(file)
        tsv_file = csv.DictReader(wrapper, delimiter="\t")
    except Exception:
        raise ValidationError("BEAST 2 log file is not a tab-separated file.")

    num_rows = sum(1 for _ in tsv_file)
    if num_rows < MIN_NUM_SNAPSHOTS:
        raise ValidationError(
            f"BEAST 2 log file should contain at least {MIN_NUM_SNAPSHOTS} entries."
        )

    buffer = file.getbuffer()
    return File(
        name=file.name,
        type=FileType.BEAST2_POSTERIOR_LOGS,
        version=1,
        size_bytes=buffer.nbytes,
        md5=md5(buffer).hexdigest(),
    )


def parse_beast2_trees(file: BytesIO) -> File:
    """Parses a BEAST 2 trees file. If the file does not seem to be a valid
    BEAST trees file, a ValidationError is raised."""
    nexus = get_nexus_from_bytesio(file)

    if not nexus.TREES or not nexus.TREES.trees:
        raise ValidationError("BEAST 2 trees contains no actual trees.")

    if len(nexus.TREES.trees) < MIN_NUM_SNAPSHOTS:
        raise ValidationError(
            f"BEAST 2 trees should contain at least {MIN_NUM_SNAPSHOTS} trees."
        )

    buffer = file.getbuffer()
    return File(
        name=file.name,
        type=FileType.BEAST2_POSTERIOR_TREES,
        version=1,
        size_bytes=buffer.nbytes,
        md5=md5(buffer).hexdigest(),
    )


def parse_other_file(file: BytesIO) -> File:
    buffer = file.getbuffer()

    # if the file contains a single tree => summary tree

    try:
        nexus = get_nexus_from_bytesio(file)
        if nexus.TREES and nexus.TREES.trees and len(nexus.TREES.trees) == 1:
            return File(
                name=file.name,
                type=FileType.SUMMARY_TREE,
                version=1,
                size_bytes=buffer.nbytes,
                md5=md5(buffer).hexdigest(),
            )
    except ValidationError:
        # file is not a valid trees file
        ...

    # otherwise we return an UNKNOWN file

    return File(
        name=file.name,
        type=FileType.UNKNOWN,
        version=1,
        size_bytes=buffer.nbytes,
        md5=md5(buffer).hexdigest(),
    )



import streamlit as st

from phylodata.data_types import Sample
from phylodata.modules.module import Module
from phylodata.parsers.parse_beast2_samples import parse_beast2_samples


class SamplesModule(Module[list[Sample]]):
    def ui(self): ...

    def validate(self): ...

    def set_dependencies(self, beast2_configuration):
        self.beast2_configuration = beast2_configuration

    def parse(self) -> list[Sample]:
        st.text(
            "Detecting the samples (e.g. using BEAST2 or language lookup). "
            "This might take a while, especially for amino acid sequences..."
        )
        return parse_beast2_samples(self.beast2_configuration)



import streamlit as st

from phylodata.data_types import FileType
from phylodata.errors import ValidationError
from phylodata.modules.module import Module
from phylodata.parsers.parse_files import parse_file


class FilesModule(Module[list]):
    def ui(self):
        with st.container(border=True):
            """
            ### Files

            Select the following files:
            """

            self.beast2_configuration = st.file_uploader(
                "BEAST 2 XML configuration", type=["xml"]
            )
            self.beast2_logs = st.file_uploader("BEAST 2 log file", type=["log"])
            self.beast2_trees = st.file_uploader(
                "BEAST 2 posterior trees", type=["trees"]
            )

            """
            Optionally, you can also select other files (like summary trees) which you consider an important part of your analysis.
            """

            self.other_files = st.file_uploader(
                "Other files (optional)", type=["*"], accept_multiple_files=True
            )

    def validate(self):
        if not self.beast2_configuration:
            raise ValidationError("Specify a BEAST 2 configuration file.")
        if not self.beast2_logs:
            raise ValidationError("Specify a BEAST 2 log file.")
        if not self.beast2_trees:
            raise ValidationError("Specify a BEAST 2 trees file.")

    def parse(self) -> list:
        if not self.beast2_configuration:
            raise ValidationError("Specify a BEAST 2 configuration file.")
        if not self.beast2_logs:
            raise ValidationError("Specify a BEAST 2 log file.")
        if not self.beast2_trees:
            raise ValidationError("Specify a BEAST 2 trees file.")

        parsed_beast2_configuration = parse_file(
            self.beast2_configuration, FileType.BEAST2_CONFIGURATION
        )
        parsed_beast2_logs = parse_file(
            self.beast2_logs, FileType.BEAST2_POSTERIOR_LOGS
        )
        parsed_beast2_trees = parse_file(
            self.beast2_trees, FileType.BEAST2_POSTERIOR_TREES
        )

        files = [
            parsed_beast2_configuration,
            parsed_beast2_logs,
            parsed_beast2_trees,
        ]

        for other_file in self.other_files:
            parsed_other_file = parse_file(other_file, FileType.UNKNOWN)
            files.append(parsed_other_file)

        return files



import taxoniq

from phylodata.data_types import ClassificationEntry, ClassificationEntryType


def look_up_taxon_classification(taxon_id: int) -> list[ClassificationEntry]:
    taxon = taxoniq.Taxon(taxon_id)
    return [
        ClassificationEntry(
            classification_id=str(t.tax_id),
            scientific_name=t.scientific_name,
            common_name=t.common_name,
            id_type=ClassificationEntryType.NCBI_TAXONOMY_ID,
        )
        for t in taxon.ranked_lineage
    ]



import streamlit as st

from phylodata.data_types import Trees
from phylodata.modules.module import Module
from phylodata.parsers.parse_trees import parse_trees


class TreesModule(Module[Trees]):
    def ui(self): ...

    def validate(self): ...

    def set_dependencies(self, beast2_trees):
        self.beast2_trees = beast2_trees

    def parse(self) -> Trees:
        st.text("Processing the trees...")
        return parse_trees(self.beast2_trees)



import bibtexparser
import streamlit as st

from phylodata.data_types import Paper
from phylodata.errors import ValidationError
from phylodata.modules.module import Module


class PaperModule(Module[Paper]):
    def ui(self):
        with st.container(border=True):
            st.write("""
            ### Paper

            Enter the following general information about your paper:
            """)

            self.title = st.text_input("Title")
            self.abstract = st.text_area("Abstract")
            self.authors = st.text_area("Authors (one per line)")
            self.bibtext = st.text_area("BibTex citation")
            self.doi = st.text_input("DOI")
            st.write(
                """Experiments with the same paper DOI will be linked to the same paper."""
            )
            self.url = st.text_input("URL (optional)")

    def validate(self):
        if not self.title.strip():
            raise ValidationError("Specify a paper title.")
        if not self.abstract.strip():
            raise ValidationError("Specify a paper abstract.")
        if not self.authors.strip():
            raise ValidationError("Specify at least one paper author.")
        if len(bibtexparser.parse_string(self.bibtext).entries) != 1:
            st.toast("Specify exactly one bibtex entry.")
            return
        if not self.doi.strip():
            st.toast("Specify a DOI.")
            return

    def parse(self) -> Paper:
        return Paper(
            title=self.title.strip(),
            authors=[author.strip() for author in self.authors.split("\n")],
            abstract=self.abstract.strip(),
            bibtex=self.bibtext.strip(),
            doi=self.doi.strip(),
            url=self.url.strip() or None,
        )



from abc import ABC, abstractmethod
from typing import Any, Optional
from xml.etree.ElementTree import ElementTree

from phylodata.data_types import ModelType, SampleType
from phylodata.utils.beast2_xml_utils import get_attribute


class Beast2PackageParser(ABC):
    """An abstract base class for parsing BEAST 2 packages."""

    def is_used(self, beast2_xml: ElementTree) -> bool:
        """Returns true if the package is detected in the given BEAST2 xml.
        Checks the top-level namespaces and the spec attribute of all elements."""
        root = beast2_xml.getroot()

        # check namespaces

        namespaces = get_attribute(root, "namespace", "")
        namespaces = namespaces.split(":")

        package_namespace = self.get_namespace()

        if namespaces and any(
            namespace == package_namespace
            or namespace.startswith(f"{package_namespace}.")
            for namespace in namespaces
        ):
            return True

        # check specs

        elements_with_spec = root.findall(".//*[@spec]")
        spec_values = [get_attribute(elem, "spec") for elem in elements_with_spec]

        for spec in spec_values:
            if not spec:
                continue

            # for the following cases: assume package_namespace = "babel.distribution"

            if spec == package_namespace:
                return True

            if spec.startswith(f"{package_namespace}."):
                # e.g. spec = "babel.distribution.Normal"
                return True

            if any(
                f"{namespace}.{spec}" == package_namespace for namespace in namespaces
            ):
                # e.g. namespace = "babel" and spec = "distribution"
                return True

            if any(
                f"{namespace}.{spec}".startswith(f"{package_namespace}.")
                for namespace in namespaces
            ):
                # e.g. namespace = "babel" and spec = "distribution.Normal"
                return True

        return False

    @abstractmethod
    def get_name(self) -> str:
        """Returns a human-readable name of the BEAST 2 package."""
        raise NotImplementedError

    @abstractmethod
    def get_description(self) -> str:
        """Returns a human-readable get_description of the BEAST 2 package."""
        raise NotImplementedError

    @abstractmethod
    def get_type(self) -> ModelType:
        """Returns the type of the BEAST 2 package."""
        raise NotImplementedError

    @abstractmethod
    def get_namespace(self) -> str:
        """Returns the namespace of the BEAST 2 package."""
        raise NotImplementedError

    @abstractmethod
    def get_documentation_url(self) -> str:
        """Returns the URL to the human-readable package website."""
        raise NotImplementedError

    def get_parameters(self, beast2_xml: ElementTree) -> dict[str, Any]:
        """Extracts potential package parameters from the BEAST 2 xml configuration."""
        return {}

    def suggest_sample_type(self) -> Optional[SampleType]:
        """If the package usage suggests a specific sample type, returns it. For example,
        a model for phylogenetic language trees would return SampleType.LANGUAGE."""
        return None


class GTRParser(Beast2PackageParser):
    def get_name(self) -> str:
        return "GTR"

    def get_description(self) -> str:
        return "General Time Reversible model of nucleotide evolution"

    def get_namespace(self) -> str:
        return "beast.evolution.substitutionmodel.GTR"

    def get_type(self) -> ModelType:
        return ModelType.SUBSTITUTION_MODEL

    def get_documentation_url(self) -> str:
        return "https://www.beast2.org/xml/beast.evolution.substitutionmodel.GTR.html"


class HKYParser(Beast2PackageParser):
    def get_name(self) -> str:
        return "HKY"

    def get_description(self) -> str:
        return "HKY85 (Hasegawa, Kishino & Yano, 1985) substitution model of nucleotide evolution"

    def get_namespace(self) -> str:
        return "beast.evolution.substitutionmodel.HKY"

    def get_type(self) -> ModelType:
        return ModelType.SUBSTITUTION_MODEL

    def get_documentation_url(self) -> str:
        return "https://www.beast2.org/xml/beast.evolution.substitutionmodel.HKY.html"


class StrictClockParser(Beast2PackageParser):
    def get_name(self) -> str:
        return "Strict Clock Model"

    def get_description(self) -> str:
        return "Defines a mean rate for each branch in the beast.tree."

    def get_namespace(self) -> str:
        return "beast.evolution.branchratemodel.StrictClockModel"

    def get_type(self) -> ModelType:
        return ModelType.CLOCK_MODEL

    def get_documentation_url(self) -> str:
        return "https://www.beast2.org/xml/beast.evolution.branchratemodel.StrictClockModel.html"


class RelaxedClockParser(Beast2PackageParser):
    def get_name(self) -> str:
        return "Relaxed Clock Model"

    def get_description(self) -> str:
        return "Defines an uncorrelated relaxed molecular clock."

    def get_namespace(self) -> str:
        return "beast.evolution.branchratemodel.UCRelaxedClockModel"

    def get_type(self) -> ModelType:
        return ModelType.CLOCK_MODEL

    def get_documentation_url(self) -> str:
        return "https://www.beast2.org/xml/beast.evolution.branchratemodel.UCRelaxedClockModel.html"


class TreeLikelihoodParser(Beast2PackageParser):
    def get_name(self) -> str:
        return "Generic TreeLikelihood"

    def get_description(self) -> str:
        return "Calculates the probability of sequence data on a tree given a site and substitution model using a variant of the 'peeling algorithm'."

    def get_namespace(self) -> str:
        return "beast.evolution.likelihood.TreeLikelihood"

    def get_type(self) -> ModelType:
        return ModelType.TREE_LIKELIHOOD

    def get_documentation_url(self) -> str:
        return (
            "https://www.beast2.org/xml/beast.evolution.likelihood.TreeLikelihood.html"
        )


class BDMMPrimeParser(Beast2PackageParser):
    def get_name(self) -> str:
        return "BDMM-Prime"

    def get_description(self) -> str:
        return "BDMM-Prime is a BEAST 2 package for performing phylodynamic inference under a variety of linear birth-death-sampling models with/without types."

    def get_namespace(self) -> str:
        return "bdmmprime"

    def get_type(self) -> ModelType:
        return ModelType.TREE_LIKELIHOOD

    def get_documentation_url(self) -> str:
        return "https://tgvaughan.github.io/BDMM-Prime/"


class BabelParser(Beast2PackageParser):
    def get_name(self) -> str:
        return "Babel"

    def get_description(self) -> str:
        return "BABEL = BEAST analysis backing effective linguistics"

    def get_namespace(self) -> str:
        return "babel"

    def get_type(self) -> ModelType:
        return ModelType.OTHER

    def get_documentation_url(self) -> str:
        return "https://github.com/rbouckaert/Babel"

    def get_parameters(self, beast2_xml: ElementTree) -> dict[str, Any]:
        return {}

    def suggest_sample_type(self) -> Optional[SampleType]:
        return SampleType.LANGUAGE


BEAST2_PACKAGE_PARSERS: list[Beast2PackageParser] = [
    GTRParser(),
    HKYParser(),
    StrictClockParser(),
    RelaxedClockParser(),
    TreeLikelihoodParser(),
    BDMMPrimeParser(),
    BabelParser(),
]
BEAST2_PACKAGE_PARSERS_PER_NAME: dict[str, Beast2PackageParser] = {
    parser.get_name(): parser for parser in BEAST2_PACKAGE_PARSERS
}



from datetime import date

import streamlit as st

from phylodata.data_types import Experiment, ExperimentType
from phylodata.modules.module import Module


class ExperimentModule(Module[Experiment]):
    def ui(self):
        with st.container(border=True):
            """
            ### Experiment

            Optionally, enter some additional information about your experiment. This is useful if you want to register multiple experiments for the same paper.
            """

            self.title = st.text_input("Experiment title (optional)")
            self.description = st.text_area("Experiment description (optional)")

    def validate(self): ...

    def parse(self) -> Experiment:
        return Experiment(
            title=self.title.strip() or None,
            description=self.description.strip() or None,
            origin="manualUpload",
            upload_date=date.today(),
            type=ExperimentType.BEAST2_Experiment,
        )



from io import BytesIO

from phylodata.parsers.parse_evolutionary_model import parse_evolutionary_model


def to_bytes_io(text: str):
    bytesio = BytesIO(bytes(text, "ascii"))
    bytesio.name = "Test"
    return bytesio


def test_detect_relaxed_clock_with_no_namespace():
    file = to_bytes_io("""<?xml version="1.0" encoding="UTF-8" standalone="no"?>
    <beast namespace="beast.core" required="BEAST.base v2.7.7" version="2.7">
        <data></data>
        <run>
            <distribution
                spec="ThreadedTreeLikelihood"
            >
                <branchRateModel clock.rate="@ucldMean.c:penguins_dna" id="RelaxedClock.c:penguins_dna" rateCategories="@rateCategories.c:penguins_dna" spec="beast.evolution.branchratemodel.UCRelaxedClockModel" tree="@Tree.t:tree">
                    <LogNormal S="@ucldStdev.c:penguins_dna" id="LogNormalDistributionModel.c:clock_dna" meanInRealSpace="true" name="distr">
                        <parameter estimate="false" id="Mrelaxed_dna" lower="0.0" name="M" upper="1.0">1.0</parameter>
                    </LogNormal>
                </branchRateModel>
            </distribution>
        </run>
    </beast>""")
    model = parse_evolutionary_model(file)

    assert model.models
    assert len(model.models) == 1
    assert model.models[0].name == "Relaxed Clock Model"



from phylodata.data_types import ClassificationEntry, DataType, Sample, SampleType
from phylodata.errors import BlastError
from phylodata.utils.blast_utils import (
    extract_taxon_ids,
    run_blast,
)
from phylodata.utils.taxon_utils import look_up_taxon_classification

MAX_SEQ_LENGTH_CONSIDERED = 160


def add_protein_metadata(samples: list[Sample]) -> list[Sample]:
    """Tries to add metadata to all samples for which we have protein
    sequences."""
    # collect all protein sequences

    protein_sequences: list[str] = []
    protein_sequence_idx: list[int] = []

    for idx, sample in enumerate(samples):
        if sample.type != SampleType.UNKNOWN:
            continue

        for data in sample.data:
            if data.type == DataType.AMINO_ACIDS:
                protein_sequences.append("".join(c for c in data.data if c.isalpha()))
                protein_sequence_idx.append(idx)
                break  # we only need one sequence per sample

    # fetch the protein classifications

    classification = fetch_protein_classifications(protein_sequences)

    # decide on the tree type

    species = set(
        [
            classification[0].scientific_name
            for classification in classification
            if classification
        ]
    )
    tree_type = SampleType.SPECIES if 1 < len(species) > 1 else SampleType.CELLS

    # add the classifications to the samples

    for idx, classification in zip(protein_sequence_idx, classification):
        if classification:
            samples[idx].classification = classification
            samples[idx].scientific_name = classification[0].scientific_name
            samples[idx].common_name = classification[0].common_name
            samples[idx].type = tree_type

    return samples


def fetch_protein_classifications(
    sequences: list[str],
) -> list[list[ClassificationEntry] | None]:
    if not sequences:
        return []

    blast_params = {
        "CMD": "Put",
        "PROGRAM": "blastp",
        "DATABASE": "nr",
        "HITLIST_SIZE": 1,
    }

    try:
        blast_results = run_blast(sequences, blast_params, MAX_SEQ_LENGTH_CONSIDERED)
    except BlastError:
        return [None] * len(sequences)

    taxon_ids = extract_taxon_ids(blast_results)

    classifications: list[list[ClassificationEntry] | None] = []

    for taxon_id in taxon_ids:
        if taxon_id:
            try:
                classifications.append(look_up_taxon_classification(taxon_id))
            except Exception:
                classifications.append(None)
        else:
            classifications.append(None)

    return classifications



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

T = TypeVar("T")


class Module(ABC, Generic[T]):
    @abstractmethod
    def ui(self):
        """Renders the streamlit UI."""
        pass

    @abstractmethod
    def validate(self):
        """Validates the input. Throws a ValidationError with a
        message if the input is invalid."""
        pass

    @abstractmethod
    def parse(self) -> T:
        """Parses the input and returns a valid object. Throws a ValidationError
        with a message if parsing fails."""
        pass



from enum import Enum

import streamlit as st

from phylodata.data_types import (
    Metadata,
    PaperWithExperiment,
)
from phylodata.errors import ValidationError
from phylodata.modules.evolutionary_model_module import EvolutionaryModelModule
from phylodata.modules.experiment_module import ExperimentModule
from phylodata.modules.files_module import FilesModule
from phylodata.modules.paper_module import PaperModule
from phylodata.modules.samples_module import SamplesModule
from phylodata.modules.trees_module import TreesModule
from phylodata.utils.output_utils import create_output_folder
from phylodata.version import __version__


class Stage(str, Enum):
    INPUT = "Input"
    NEXT_STEPS = "Next Steps"


STAGE = "stage"
OUTPUT_FOLDER = "output_folder"

paper_module = PaperModule()
experiment_module = ExperimentModule()
files_module = FilesModule()
evolutionary_model_module = EvolutionaryModelModule()
samples_module = SamplesModule()
trees_module = TreesModule()

if STAGE not in st.session_state:
    st.session_state[STAGE] = Stage.INPUT


if st.session_state[STAGE] == Stage.INPUT:
    """
    # 🌴 Process Experiment

    Let's process your Bayesian Phylogenetics experiment to prepare it for upload to the PhyloData repository.

    Note that right now, we only support BEAST 2 experiments.

    This usually takes less than five minutes.
    """

    paper_module.ui()
    experiment_module.ui()
    files_module.ui()

    """
    Almost done!
    """

    def process_input():
        with st.spinner("Processing your analysis..."):
            try:
                paper_module.validate()
                files_module.validate()

                paper = paper_module.parse()
                experiment = experiment_module.parse()

                st.text("Processing the given files...")
                files = files_module.parse()

                evolutionary_model_module.set_dependencies(
                    files_module.beast2_configuration
                )
                parsed_evolutionary_model = evolutionary_model_module.parse()

                samples_module.set_dependencies(files_module.beast2_configuration)
                parsed_samples = samples_module.parse()

                trees_module.set_dependencies(files_module.beast2_trees)
                parsed_trees = trees_module.parse()

            except ValidationError as error:
                st.toast(error.message)
                return

            metadata = Metadata(evo_data_pipeline_version=__version__)

            paper_with_experiment = PaperWithExperiment(
                experiment=experiment,
                paper=paper,
                files=files,
                evolutionary_model=parsed_evolutionary_model,
                samples=parsed_samples,
                trees=parsed_trees,
                metadata=metadata,
            )

            st.text("Creating PhyloData folder...")

            output_folder = create_output_folder(
                paper.title,
                files_module.beast2_configuration,  # type: ignore
                files_module.beast2_logs,  # type: ignore
                files_module.beast2_trees,  # type: ignore
                files_module.other_files,
                paper_with_experiment,
            )

            st.session_state[STAGE] = Stage.NEXT_STEPS
            st.session_state[OUTPUT_FOLDER] = output_folder
            st.rerun()

    process_experiment_button = st.button("Process experiment", type="primary")
    if process_experiment_button:
        process_input()


if st.session_state[STAGE] == Stage.NEXT_STEPS:
    """
    # 🚀 Next Steps

    You successfully processed your experiment! These are your next steps:

    1. Create a ZIP file of the newly created folder.
    2. Upload it on the [PhyloData website](http://phylodata.com/new-experiment).
    """



import os
import shutil
from io import BytesIO

import msgspec
from streamlit.runtime.uploaded_file_manager import UploadedFile

from phylodata.data_types import PaperWithExperiment

WASABI_BUCKET_NAME = "phylodata-experiments"


def create_output_folder(
    title: str,
    beast2_configuration: BytesIO,
    beast2_logs: BytesIO,
    beast2_trees: BytesIO,
    other_files: list[UploadedFile],
    paper_with_experiment: PaperWithExperiment,
) -> str:
    """
    Creates a folder with a name based on the given title and writes the provided
    files and the PhyloData metadata into it.
    """
    output_folder = f"{title}-phylodata"

    if os.path.exists(output_folder):
        shutil.rmtree(output_folder)
    os.mkdir(output_folder)

    with open(f"{output_folder}/phylodata_metadata.json", "wb") as f:
        f.write(
            msgspec.json.format(msgspec.json.encode(paper_with_experiment), indent=2)
        )

    used_filenames = set("phylodata_metadata.json")
    all_files = [beast2_configuration, beast2_logs, beast2_trees] + other_files

    for file in all_files:
        filename = file.name
        base, ext = os.path.splitext(file.name)

        while filename in used_filenames:
            filename = f"{base}_{len(used_filenames)}{ext}"

        used_filenames.add(filename)
        with open(f"{output_folder}/{filename}", "wb") as f:
            f.write(file.getbuffer())

    return output_folder



"""This file contains all the types which are used to describe a PhyloData experiment."""

from datetime import date
from enum import Enum
from typing import Any, Optional

import msgspec


class ExperimentType(Enum):
    BEAST2_Experiment = "beast2Experiment"


class ClassificationEntryType(Enum):
    NCBI_TAXONOMY_ID = "ncibTaxonomyId"
    GLOTTOLOG_ID = "glottologId"


class FileType(Enum):
    BEAST2_CONFIGURATION = "beast2Configuration"
    BEAST2_POSTERIOR_LOGS = "beast2PosteriorLogs"
    BEAST2_POSTERIOR_TREES = "beast2PosteriorTrees"
    SUMMARY_TREE = "summaryTree"
    CODEPHY_MODEL = "codephyModel"
    EVO_DATA_EXPERIMENT = "phyloDataExperiment"
    UNKNOWN = "unknown"


class SampleType(Enum):
    SPECIES = "species"
    CELLS = "cells"
    LANGUAGE = "language"
    UNKNOWN = "unknown"


class DataType(Enum):
    RNA = "rna"
    DNA = "dna"
    AMINO_ACIDS = "aminoAcids"
    PHASED_DIPLOID_DNA = "phasedDiploidDna"
    TRAITS = "traits"
    UNKNOWN = "unknown"


class ModelType(Enum):
    SUBSTITUTION_MODEL = "substitutionModel"
    CLOCK_MODEL = "clockModel"
    TREE_PRIOR = "treePrior"
    TREE_LIKELIHOOD = "treeLikelihood"
    OTHER = "other"


class Experiment(msgspec.Struct, rename="camel"):
    type: ExperimentType
    origin: str
    upload_date: date
    title: Optional[str] = None
    description: Optional[str] = None
    license: str = "CC0"
    id: Optional[str] = None


class Paper(msgspec.Struct, rename="camel"):
    doi: str
    title: str
    authors: list[str]
    abstract: str
    bibtex: str
    url: Optional[str] = None


class File(msgspec.Struct, rename="camel"):
    name: str
    type: FileType
    version: int
    size_bytes: int
    md5: str
    local_path: Optional[str] = None
    remote_path: Optional[str] = None


class SampleData(msgspec.Struct, rename="camel"):
    type: DataType
    length: int
    data: str


class ClassificationEntry(msgspec.Struct, rename="camel"):
    classification_id: str
    scientific_name: str
    id_type: ClassificationEntryType
    common_name: Optional[str] = None
    id: Optional[str] = None


class Sample(msgspec.Struct, rename="camel"):
    sample_id: str
    scientific_name: str
    type: SampleType
    classification: list[ClassificationEntry]
    data: list[SampleData]
    common_name: Optional[str] = None
    id: Optional[str] = None


class Trees(msgspec.Struct, rename="camel"):
    number_of_trees: int
    number_of_tips: int
    ultrametric: bool
    time_tree: bool
    rooted: bool
    ccd1_entropy: float
    tree_ess: int
    ccd0_map_tree: str
    hipstr_tree: str
    leaf_to_sample_map: dict[str, str]
    average_root_age: float


class EvolutionaryModelComponent(msgspec.Struct, rename="camel"):
    name: str
    description: str
    type: ModelType
    documentation_url: str
    parameters: dict[str, Any]


class EvolutionaryModel(msgspec.Struct, rename="camel"):
    models: list[EvolutionaryModelComponent]


class Metadata(msgspec.Struct, rename="camel"):
    evo_data_pipeline_version: str


class PaperWithExperiment(msgspec.Struct, rename="camel"):
    experiment: Experiment
    paper: Paper
    files: list[File]
    samples: list[Sample]
    trees: Trees
    evolutionary_model: EvolutionaryModel
    metadata: Metadata


def validate_json(file_path: str):
    """Validates if the given file contains a valid PaperWithExperiment object.
    Throws a msgspec.ValidationError if not."""
    with open(file_path) as handle:
        msgspec.json.decode(handle.read(), type=PaperWithExperiment)


def get_schema():
    """Returns the formatted JSON schema corresponding to PaperWithExperiment."""
    schema = msgspec.json.schema(PaperWithExperiment)
    return msgspec.json.format(msgspec.json.encode(schema).decode())
